from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.strategies import integers
from pytest import raises

from utilities.hypothesis import git_repos, settings_with_reduced_examples
from utilities.python_dotenv import LoadSettingsError, load_settings

if TYPE_CHECKING:
    from pathlib import Path


class TestLoadSettings:
    @given(root=git_repos(), value=integers())
    @settings_with_reduced_examples()
    def test_main(self, *, root: Path, value: int) -> None:
        @dataclass(frozen=True)
        class Settings:
            key: str

        with root.joinpath(".env").open(mode="w") as fh:
            _ = fh.write(f"key = {value}\n")

        settings = load_settings(Settings, cwd=root)
        expected = Settings(key=str(value))
        assert settings == expected

    @given(root=git_repos(), value=integers())
    @settings_with_reduced_examples()
    def test_upper_case_dotenv(self, *, root: Path, value: int) -> None:
        @dataclass(frozen=True)
        class Settings:
            key: str

        with root.joinpath(".env").open(mode="w") as fh:
            _ = fh.write(f"KEY = {value}\n")

        settings = load_settings(Settings, cwd=root)
        expected = Settings(key=str(value))
        assert settings == expected

    @given(root=git_repos(), value=integers())
    @settings_with_reduced_examples()
    def test_upper_case_key(self, *, root: Path, value: int) -> None:
        @dataclass(frozen=True)
        class Settings:
            KEY: str

        with root.joinpath(".env").open(mode="w") as fh:
            _ = fh.write(f"key = {value}\n")

        settings = load_settings(Settings, cwd=root)
        expected = Settings(KEY=str(value))
        assert settings == expected

    @given(root=git_repos(), value=integers())
    @settings_with_reduced_examples()
    def test_extra_key(self, *, root: Path, value: int) -> None:
        @dataclass(frozen=True)
        class Settings:
            key: str

        with root.joinpath(".env").open(mode="w") as fh:
            _ = fh.write(f"key = {value}\n")
            _ = fh.write(f"other = {value}\n")

        settings = load_settings(Settings, cwd=root)
        expected = Settings(key=str(value))
        assert settings == expected

    @given(root=git_repos())
    @settings_with_reduced_examples()
    def test_error_file_not_found(self, *, root: Path) -> None:
        @dataclass(frozen=True)
        class Settings:
            KEY: str

        with raises(LoadSettingsError, match=r"Path '.*' must exist\."):
            _ = load_settings(Settings, cwd=root)

    @given(root=git_repos())
    @settings_with_reduced_examples()
    def test_error_field_missing(self, *, root: Path) -> None:
        @dataclass(frozen=True)
        class Settings:
            key: str

        root.joinpath(".env").touch()

        with raises(LoadSettingsError, match=r"Field 'key' must exist\."):
            _ = load_settings(Settings, cwd=root)

    @given(root=git_repos(), value=integers())
    @settings_with_reduced_examples()
    def test_error_field_duplicated(self, *, root: Path, value: int) -> None:
        @dataclass(frozen=True)
        class Settings:
            key: str

        with root.joinpath(".env").open(mode="w") as fh:
            _ = fh.write(f"key = {value}\n")
            _ = fh.write(f"KEY = {value}\n")

        with raises(
            LoadSettingsError, match=r"Field 'key' must exist exactly once; got .*\."
        ):
            _ = load_settings(Settings, cwd=root)



from __future__ import annotations

from typing import TYPE_CHECKING

from pytest import mark, param

from utilities.sentinel import SENTINEL_REPR, Sentinel, sentinel

if TYPE_CHECKING:
    from collections.abc import Callable


class TestSentinel:
    def test_isinstance(self) -> None:
        assert isinstance(sentinel, Sentinel)

    @mark.parametrize("method", [param(repr), param(str)])
    def test_repr_and_str(self, method: Callable[..., str]) -> None:
        assert method(sentinel) == SENTINEL_REPR

    def test_singletone(self) -> None:
        assert Sentinel() is sentinel



from __future__ import annotations

from dataclasses import dataclass, fields
from typing import TYPE_CHECKING, TypeVar

from dotenv import dotenv_values
from typing_extensions import override

from utilities.dataclasses import Dataclass
from utilities.git import get_repo_root
from utilities.iterables import (
    _OneStrCaseInsensitiveBijectionError,
    _OneStrCaseInsensitiveEmptyError,
    one_str,
)
from utilities.pathlib import PWD

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping
    from pathlib import Path

    from utilities.types import PathLike

_TDataclass = TypeVar("_TDataclass", bound=Dataclass)


def load_settings(cls: type[_TDataclass], /, *, cwd: PathLike = PWD) -> _TDataclass:
    """Load a set of settings from the `.env` file."""
    path = get_repo_root(cwd=cwd).joinpath(".env")
    if not path.exists():
        raise _LoadSettingsFileNotFoundError(path=path) from None
    maybe_values = dotenv_values(path)
    values = {k: v for k, v in maybe_values.items() if v is not None}

    def yield_items() -> Iterator[tuple[str, str]]:
        for fld in fields(cls):
            try:
                key = one_str(values, fld.name, case_sensitive=False)
            except _OneStrCaseInsensitiveEmptyError:
                raise _LoadSettingsEmptyError(path=path, field=fld.name) from None
            except _OneStrCaseInsensitiveBijectionError as error:
                raise _LoadSettingsNonUniqueError(
                    path=path, field=fld.name, counts=error.counts
                ) from None
            else:
                yield fld.name, values[key]

    return cls(**dict(yield_items()))


@dataclass(kw_only=True)
class LoadSettingsError(Exception):
    path: Path


@dataclass(kw_only=True)
class _LoadSettingsFileNotFoundError(LoadSettingsError):
    @override
    def __str__(self) -> str:
        return f"Path {str(self.path)!r} must exist."


@dataclass(kw_only=True)
class _LoadSettingsEmptyError(LoadSettingsError):
    field: str

    @override
    def __str__(self) -> str:
        return f"Field {self.field!r} must exist."


@dataclass(kw_only=True)
class _LoadSettingsNonUniqueError(LoadSettingsError):
    field: str
    counts: Mapping[str, int]

    @override
    def __str__(self) -> str:
        return f"Field {self.field!r} must exist exactly once; got {self.counts}."


__all__ = ["LoadSettingsError", "load_settings"]



from __future__ import annotations

from asyncio import sleep
from inspect import signature
from typing import Annotated, Any

from pytest import raises

from tests.conftest import FLAKY
from utilities.atools import (
    RefreshMemoizedError,
    _memoize_auto_keygen_is_param,
    memoize,
    no_memoize,
    refresh_memoized,
)


class TestMemoize:
    async def test_main(self) -> None:
        i = 0

        @memoize
        async def increment() -> int:
            nonlocal i
            i += 1
            return i

        for _ in range(2):
            assert (await increment()) == 1

    @FLAKY
    async def test_with_duration(self) -> None:
        i = 0

        @memoize(duration=0.01)
        async def increment() -> int:
            nonlocal i
            i += 1
            return i

        for _ in range(2):
            assert (await increment()) == 1
        await sleep(0.01)
        for _ in range(2):
            assert (await increment()) == 2

    async def test_with_keygen(self) -> None:
        i = 0

        @memoize
        async def increment(j: int, /, *, ignore: Annotated[bool, no_memoize]) -> int:
            nonlocal i
            i += j
            _ = ignore
            return i

        for j in [True, False]:
            assert (await increment(1, ignore=j)) == 1
        for j in [True, False]:
            assert (await increment(2, ignore=j)) == 3


class TestMemoizeAutoKeygenIsParam:
    def test_no_annotation(self) -> None:
        def func(a, /) -> Any:  # noqa: ANN001 # pyright: ignore[reportMissingParameterType]
            return a

        ann = signature(func).parameters["a"].annotation
        result = _memoize_auto_keygen_is_param(ann)
        assert result is True

    def test_basic_annotation(self) -> None:
        def func(a: int, /) -> int:
            return a

        ann = signature(func).parameters["a"].annotation
        result = _memoize_auto_keygen_is_param(ann)
        assert result is True

    def test_no_memoize(self) -> None:
        def func(a: Annotated[int, no_memoize], /) -> int:
            return a

        ann = signature(func).parameters["a"].annotation
        result = _memoize_auto_keygen_is_param(ann)
        assert result is False


class TestRefreshMemoized:
    @FLAKY
    async def test_main(self) -> None:
        i = 0

        @memoize(duration=0.01)
        async def increment() -> int:
            nonlocal i
            i += 1
            return i

        for _ in range(2):
            assert (await increment()) == 1
        await sleep(0.01)
        for _ in range(2):
            assert (await increment()) == 2
        assert await refresh_memoized(increment) == 3

    async def test_error(self) -> None:
        async def none() -> None:
            return None

        with raises(
            RefreshMemoizedError, match="Asynchronous function .* must be memoized"
        ):
            await refresh_memoized(none)
